#    Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

DEFAULT = {
    'arp_disable': False,
    'default_virtual_server_vrid': None,
    'logging_template': None,
    'policy_template': None,
    'template_virtual_server': None,
    'default_virtual_server_vrid': None,
    'ipinip': False,
    'no_dest_nat': None,
    'ha_conn_mirror': None,
    'template_virtual_port': None,
    'template_tcp': None,
    'template_policy': None,
    'autosnat': True,
    'conn_limit': None,
    'template_http': None,
    'templates': None,
    'conn_limit': None,
    'conn_resume': None,
    'templates': None,
    'use_rcv_hop_for_resp': False
}
